/* idrimg.cpp--Idrimg class and member function definitions */

#include <windows.h>
#include <malloc.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include "idrimg.h"

bool GetVal(char * input, char * keystr, char * valstr);

Idrimg::Idrimg(char * img_path) {

   char fpath[_MAX_PATH];
   int epos;
   FILE *infile;
   char input[81];
   char temp[81];
   char valstr[81];

   // Get .IMG path and try to open .DOC file

   Valid = false;
   DataType = IMG_NODTYPE;
   FileType = IMG_NOFTYPE;
   strcpy (Path, img_path);
   epos = strrchr (Path, '.') - Path + 1;
   strcpy(fpath,Path);
   strncpy(fpath + epos, "doc", 3);
   if ((infile = fopen(fpath,"r")) == NULL)
   {
      return;
   }

   /* read DOC file */

   while (fgets(input, 80, infile) != NULL)
   {
      strcpy(temp,"data type");
      if (GetVal(input,temp,valstr))
      {
         if (strcmp(valstr,"byte") == 0)
            DataType = IMG_BYTE;
         if (strcmp(valstr,"integer") == 0)
            DataType = IMG_INTEGER;
         if (strcmp(valstr,"real") == 0)
            DataType = IMG_REAL;
         continue;
      }
      strcpy(temp,"file type");
      if (GetVal(input,temp,valstr))
      {
         if (strcmp(valstr,"ascii") == 0)
            FileType = IMG_ASCII;
         if (strcmp(valstr,"binary") == 0)
            FileType = IMG_BINARY;
         if (strcmp(valstr,"packed binary") == 0)
            FileType = IMG_PACKED;
         continue;
      }
      strcpy(temp,"columns");
      if (GetVal(input,temp,valstr))
      {
         Cols = atoi(valstr);
         continue;
      }
      strcpy(temp,"rows");
      if (GetVal(input,temp,valstr))
      {
         Rows = atoi(valstr);
         continue;
      }
      strcpy(temp,"min. X");
      if (GetVal(input,temp,valstr))
      {
         MinX = atof(valstr);
         continue;
      }
      strcpy(temp,"min. Y");
      if (GetVal(input,temp,valstr))
      {
         MinY = atof(valstr);
         continue;
      }
      strcpy(temp,"max. X");
      if (GetVal(input,temp,valstr))
      {
         MaxX = atof(valstr);
         continue;
      }
      strcpy(temp,"max. Y");
      if (GetVal(input,temp,valstr))
      {
         MaxY = atof(valstr);
         continue;
      }
      strcpy(temp,"min. value");
      if (GetVal(input,temp,valstr))
      {
         MinVal = atof(valstr);
         continue;
      }
      strcpy(temp,"max. value");
      if (GetVal(input,temp,valstr))
      {
         MaxVal = atof(valstr);
         continue;
      }
   }
   fclose(infile);

   // Check for existence of .PAL file

   Palette = IMG_NOPALETTE;
   strncpy(fpath + epos, "pal", 3);
   if ((infile = fopen(fpath,"r")) != NULL)
   {
      Palette = IMG_PAL;
      fclose(infile);
   };
   strncpy(fpath + epos, "smp", 3);
   if ((infile = fopen(fpath,"rb")) != NULL)
   {
      Palette = IMG_SMP;
      fclose(infile);
   };

   // IMG_ASCII not supported

   if (FileType != IMG_ASCII) Valid = true;

   return;
};

int Idrimg::IsValid() {

   return Valid;

};

int Idrimg::Query(long * nbands, long * width, long * height,
   double * mbr) {

   if (! Valid) return false;

   long mb = 1;

   *nbands = mb;
   *width = Cols;
   *height = Rows;
   mbr[0] = MinX;
   mbr[1] = MaxY;
   mbr[2] = MaxX;
   mbr[3] = MinY;

   return true;

};

int Idrimg::WriteBIP(char * out_path, double * clip_mbr,
   long out_width, long out_height) {

   size_t NROWS, NCOLS;
   double ULXMAP, ULYMAP, XDIM, YDIM;
   double iulx, iuly, ixdim, iydim;
   double oxmin, oymin, oxmax, oymax;
   int epos, result, data_width;
   char pal_path[_MAX_PATH];
   char bip_path[_MAX_PATH];
   char hdr_path[_MAX_PATH];
   char clr_path[_MAX_PATH];
   FILE *in_img, *in_pal, *out_bip, *out_hdr, *out_clr;
   bool success, wr_degen, done, HasPalette;

   unsigned char *in_byte, bbuf[1024], smp[786], *out_byte;
   short *in_integer, ibuf[1024];
   float *in_real;

   double sf1, sf2, tmp;
   fpos_t f_pos, v_pos;
   size_t f_read, f_write;
   size_t bs, c0, i0, j0, k0, i1, j1, t0, ip;
   char input[81];
   char *tok;
   int val, r, g, b, smin, smax;

   // Debugging variables

   bool debug = false;
   char log_path[_MAX_PATH];
   FILE *out_log;

   if (debug)
   {
      strcpy(log_path,out_path);
      epos = strrchr(log_path, '.') - log_path + 1;
      strncpy(log_path + epos, "log", 3);
      success = (! ((out_log = fopen(log_path,"w")) == NULL));
      if (! success) return false;
      fprintf(out_log,"Path = '%s'\n",Path);
      fprintf(out_log,"out_path = '%s'\n",out_path);
      fprintf(out_log,"clip_mbr[0] = '%f'\n",clip_mbr[0]);
      fprintf(out_log,"clip_mbr[1] = '%f'\n",clip_mbr[1]);
      fprintf(out_log,"clip_mbr[2] = '%f'\n",clip_mbr[2]);
      fprintf(out_log,"clip_mbr[3] = '%f'\n",clip_mbr[3]);
      fprintf(out_log,"out_width = '%i'\n",out_width);
      fprintf(out_log,"out_height = '%i'\n",out_height);
   }

   // Check for valid image

   if (! Valid)  return false;

   // Calculate image parameters

   ixdim = (MaxX - MinX) / Cols;
   iydim = (MaxY - MinY) / Rows;
   iulx = MinX + (ixdim / 2);
   iuly = MaxY - (iydim / 2);
   sf1 = MaxVal - MinVal;

   // If initializing image or clip_mbr is completely outside
   // write a degenerate image (pixel 0,0)

   wr_degen = false;
   if ((clip_mbr[0] > MaxX) || (clip_mbr[2] < MinX))
      wr_degen = true;
   if ((clip_mbr[3] > MaxY) || (clip_mbr[1] < MinY))
      wr_degen = true;
   if (out_width == 0) wr_degen = true;
   if (wr_degen)
   {

      // Set degenerate parameters

      ULXMAP = iulx;
      ULYMAP = iuly;
      XDIM = ixdim;
      YDIM = iydim;
      NCOLS = 1;
      NROWS = 1;
   }
   else
   {
      oxmin = __max((clip_mbr[0] - ixdim), MinX);
      oymin = __max((clip_mbr[3] - iydim), MinY);
      oxmax = __min((clip_mbr[2] + ixdim), MaxX);
      oymax = __min((clip_mbr[1] + iydim), MaxY);
      XDIM = (clip_mbr[2] - clip_mbr[0]) / out_width;
      if (XDIM < ixdim)
      {
         XDIM = ixdim;
         tmp = floor((oxmin - MinX) / ixdim);
         oxmin = tmp * ixdim + MinX;
      }
      YDIM = (clip_mbr[1] - clip_mbr[3]) / out_height;
      if (YDIM < iydim)
      {
         YDIM = iydim;
         tmp = floor((MaxY - oymax) / iydim);
         oymax = MaxY - (tmp * iydim);
      }
      ULXMAP = oxmin + (XDIM / 2);
      ULYMAP = oymax - (YDIM / 2);
      NCOLS = (double) (oxmax - oxmin) / XDIM + 0.5;
      NROWS = (double) (oymax - oymin) / YDIM + 0.5;
   }

   // Allocate memory for I/O buffers

   c0 = (double) (ULXMAP - iulx) / ixdim + 0.5;
   j1 = (double) ((ULXMAP - iulx) + ((NCOLS - 1) * XDIM)) / ixdim + 0.5;
   bs = j1 - c0 + 1;
   if (debug)
   {
      fprintf(out_log,"c0 = '%i'\n",c0);
      fprintf(out_log,"j1 = '%i'\n",j1);
      fprintf(out_log,"bs = '%i'\n",bs);
   }
   switch (DataType)
   {
      case IMG_BYTE:
         data_width = 1;
         in_byte = (unsigned char *) calloc(bs,1);
         if (in_byte == NULL) return false;
         break;
      case IMG_INTEGER:
         data_width = 2;
         in_integer = (short *) calloc(bs,2);
         if (in_integer == NULL) return false;
         break;
      case IMG_REAL:
         data_width = 4;
         in_real = (float *) calloc(bs,4);
         if (in_real == NULL) return false;
         break;
   }
   out_byte = (unsigned char *) calloc(NCOLS,1);
   if (out_byte == NULL) return false;

   // establish file paths

   strcpy(bip_path,out_path);
   epos = strrchr(bip_path, '.') - bip_path + 1;
   strcpy(hdr_path,bip_path);
   strncpy(hdr_path + epos, "hdr", 3);
   HasPalette = false;
   switch (Palette)
   {
      case IMG_PAL:
         epos = strrchr(Path, '.') - Path + 1;
         strcpy(pal_path,Path);
         strncpy(pal_path + epos, "pal", 3);
         HasPalette = true;
         break;
      case IMG_SMP:
         epos = strrchr(Path, '.') - Path + 1;
         strcpy(pal_path,Path);
         strncpy(pal_path + epos, "smp", 3);
         HasPalette = true;
         break;
   }

   // Open input and output files

   success = true;
   success = success && (! ((in_img = fopen(Path,"rb")) == NULL));
   success = success && (! ((out_bip = fopen(bip_path,"wb")) == NULL));
   success = success && (! ((out_hdr = fopen(hdr_path,"w")) == NULL));
   if (HasPalette)
   {
      epos = strrchr(bip_path, '.') - bip_path + 1;
      strcpy(clr_path,bip_path);
      strncpy(clr_path + epos, "clr", 3);

      switch (Palette)
      {
         case IMG_PAL:
            success = success
               && (! ((in_pal = fopen(pal_path,"r")) == NULL));
            break;
         case IMG_SMP:
            success = success
               && (! ((in_pal = fopen(pal_path,"rb")) == NULL));
            break;
      }
      success = success
         && (! ((out_clr = fopen(clr_path,"w")) == NULL));
   }
   if (! success)
   {
      _fcloseall();
      result = remove(bip_path);
      result = remove(hdr_path);
      result = remove(clr_path);
      return false;
   }

   // Write header file

   fprintf(out_hdr,"BYTEORDER I\n");
   fprintf(out_hdr,"LAYOUT BIP\n");
   fprintf(out_hdr,"NROWS %i\n", NROWS);
   fprintf(out_hdr,"NCOLS %i\n", NCOLS);
   fprintf(out_hdr,"NBANDS 1\n");
   fprintf(out_hdr,"NBITS 8\n");
   fprintf(out_hdr,"ULXMAP %f\n", ULXMAP);
   fprintf(out_hdr,"ULYMAP %f\n", ULYMAP);
   fprintf(out_hdr,"XDIM %f\n", XDIM);
   fprintf(out_hdr,"YDIM %f\n", YDIM);

   // If HasPalette then write colormap file from palette

   switch (Palette)
   {
      case IMG_PAL:
         while (fgets(input, 80, in_pal) != NULL)
         {
            tok = strtok(input, " \n");
            val = atoi(tok);
            tok = strtok(NULL, " \n");
            r = (float) atoi(tok) / 63 * 255 + 0.5;
            tok = strtok(NULL, " \n");
            g = (float) atoi(tok) / 63 * 255 + 0.5;
            tok = strtok(NULL, " \n");
            b = (float) atoi(tok) / 63 * 255 + 0.5;
            fprintf(out_clr,"%i %i %i %i\n", val, r, g, b);
         }
         smin = 0;
         smax = 255;
         sf2 = 255;
         break;
      case IMG_SMP:
         f_read = fread(smp,1,786,in_pal);
         for (int v = 0; v < 256; v++)
         {
            r = smp[v * 3 + 18];
            g = smp[v * 3 + 19];
            b = smp[v * 3 + 20];
            fprintf(out_clr,"%i %i %i %i\n", v, r, g, b);
         }
         smin = smp[14];
         smax = smp[16];
         sf2 = smax - smin;
         break;
   }

   // For each output cell, find nearest neighbor in Idrimg

   i0 = -1;
   v_pos = -1;
   j0 = -1;
   k0 = -1;
   for (size_t i = 0; i < NROWS; i++)
   {
      i1 = (double) (iuly - ULYMAP + (i * YDIM)) / iydim + 0.5;
      if (i1 != i0)
      {

         // Begin new row

         i0 = i1;
         f_pos = (i0 * Cols + c0) * data_width;

         switch (FileType)
         {
            case IMG_BINARY:

               // Set the file pointer and read in the row

               result = fsetpos(in_img, &f_pos);
               switch (DataType)
               {
                  case IMG_BYTE:
                     f_read = fread(in_byte,data_width,bs,in_img);
                     break;
                  case IMG_INTEGER:
                     f_read = fread(in_integer,data_width,bs,in_img);
                     break;
                  case IMG_REAL:
                     f_read = fread(in_real,data_width,bs,in_img);
                     break;
               }
               if ((f_read != bs) && debug)
               {
                  fprintf(out_log,"Could not read %i bytes at %i\n",bs,f_pos);
                  fprintf(out_log,"f_read = '%i'\n",bs,f_read);
                  _fcloseall();
                  return false;
               }
               break;

            case IMG_PACKED:

               // Read in values until the input row is generated

               ip = -1;
               done = false;
               do
               {
                  if (j0 == -1)
                  {

                     // IMG_REAL images are never packed

                     if (DataType == IMG_BYTE)
                        f_read = fread(bbuf,data_width,1024,in_img);
                     else
                        f_read = fread(ibuf,data_width,1024,in_img);
                     j0 = 0;
                     k0 = 0;
                     f_read = f_read / 2;
                  }
                  t0 = j0 / 2;
                  for (size_t j = t0; j < f_read; j++)
                  {
                     j0 = j + j;
                     if (DataType == IMG_BYTE)
                        j1 = bbuf[j0] - k0;
                     else
                        j1 = ibuf[j0] - k0;
                     k0 = 0;
                     if ((v_pos + j1) < f_pos)
                     {
                        v_pos = v_pos + j1;
                        continue;
                     }
                     for (size_t k = 1; k <= j1; k++)
                     {
                        if ((v_pos + k) < f_pos)
                        {
                           continue;
                        }
                        ip++;
                        if (DataType == IMG_BYTE)
                           in_byte[ip] = bbuf[j0 + 1];
                        else
                           in_integer[ip] = ibuf[j0 + 1];
                        if (ip == (bs - 1))
                        {
                           k0 = k;
                           break;
                        }
                     }
                     if (ip == (bs - 1))
                     {
                        v_pos = v_pos + k0;
                        break;
                     }
                     v_pos = v_pos + j1;
                  }
                  if (v_pos < (f_pos + bs - 1))
                  {
                     j0 = -1;
                  }
                  else
                     done = true;
                  if ((f_read < 512) && (! done))
                  {
                     if (debug)
                     {
                        fprintf(out_log,"Could not read %i bytes at %i\n",bs,f_pos);
                        fprintf(out_log,"f_read = '%i'\n",bs,f_read);
                        _fcloseall();
                        return false;
                     }
                     break;
                  }
               }
               while  (! done);
               break;
         }

         for (size_t j = 0; j < NCOLS; j++)
         {
            j1 = (double) ((ULXMAP - iulx) + (j * XDIM)) / ixdim + 0.5;
            j1 = j1 - c0;
            switch (DataType)
            {
               case IMG_BYTE:
                  out_byte[j] = in_byte[j1];
                  break;
               case IMG_INTEGER:
                  if (HasPalette)
                     out_byte[j] = (float) (in_integer[j1] - MinVal) / sf1 * sf2 + smin;
                  else
                     out_byte[j] = (float) (in_integer[j1] - MinVal) / sf1 * 255;
                  break;
               case IMG_REAL:
                  if (HasPalette)
                     out_byte[j] = (float) (in_real[j1] - MinVal) / sf1 * sf2 + smin;
                  else
                     out_byte[j] = (float) (in_real[j1] - MinVal) / sf1 * 255;
                  break;
            }
         }
      }

      // Write out row

      f_write = fwrite(out_byte,1,NCOLS,out_bip);
   }

   // Clean up I/O buffers

   switch (DataType)
   {
      case IMG_BYTE:
         free (in_byte);
         break;
      case IMG_INTEGER:
         free (in_integer);
         break;
      case IMG_REAL:
         free (in_real);
         break;
   }
   free (out_byte);

   // Let's get the heck outta here

   _fcloseall();
   return true;

};

bool GetVal(char * input, char * keystr, char * valstr)
{
   int tmp, bpos, epos;
   tmp = strlen(keystr);
   if (strncmp(input,keystr,tmp) != 0) return false;
   bpos = strchr (input, ':') - input + 2;
   epos = strchr (input, '\n') - input;
   strncpy (valstr, input + bpos, epos - bpos);
   valstr[epos - bpos] = '\0';
   return true;
};

